import hashlib
import os
import platform
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import traceback
import venv

from clarifai_grpc.grpc.api import resources_pb2, service_pb2

from clarifai.runners.models.model_builder import ModelBuilder
from clarifai.utils.logging import logger


class ModelRunLocally:
    def __init__(self, model_path):
        self.model_path = model_path
        self.requirements_file = os.path.join(self.model_path, "requirements.txt")

        # ModelBuilder contains multiple useful methods to interact with the model
        self.builder = ModelBuilder(self.model_path, download_validation_only=True)
        self.config = self.builder.config

    def _requirements_hash(self):
        """Generate a hash of the requirements file."""
        with open(self.requirements_file, "r") as f:
            return hashlib.md5(f.read().encode('utf-8')).hexdigest()

    def _get_env_executable(self):
        """Get the python executable from the virtual environment."""
        # Depending on the platform, venv scripts are placed in either "Scripts" (Windows) or "bin" (Linux/Mac)
        if platform.system().lower().startswith("win"):
            scripts_folder = "Scripts"
            python_exe = "python.exe"
            pip_exe = "pip.exe"
        else:
            scripts_folder = "bin"
            python_exe = "python"
            pip_exe = "pip"

        self.python_executable = os.path.join(self.venv_dir, scripts_folder, python_exe)
        self.pip_executable = os.path.join(self.venv_dir, scripts_folder, pip_exe)

        return self.python_executable, self.pip_executable

    def create_temp_venv(self):
        """Create a temporary virtual environment."""
        requirements_hash = self._requirements_hash()

        temp_dir = os.path.join(tempfile.gettempdir(), str(requirements_hash))
        venv_dir = os.path.join(temp_dir, "venv")

        if os.path.exists(temp_dir):
            logger.info(f"Using previous virtual environment at {temp_dir}")
            use_existing_venv = True
        else:
            logger.info("Creating temporary virtual environment...")
            use_existing_venv = False
            venv.create(venv_dir, with_pip=True)
            logger.info(f"Created temporary virtual environment at {venv_dir}")

        self.venv_dir = venv_dir
        self.temp_dir = temp_dir
        self.python_executable, self.pip_executable = self._get_env_executable()

        return use_existing_venv

    def install_requirements(self):
        """Install the dependencies from requirements.txt and Clarifai."""
        _, pip_executable = self._get_env_executable()
        try:
            logger.info(
                f"Installing requirements from {self.requirements_file}... in the virtual environment {self.venv_dir}"
            )
            subprocess.check_call([pip_executable, "install", "-r", self.requirements_file])
            logger.info("Installing Clarifai package...")
            subprocess.check_call([pip_executable, "install", "clarifai"])
            logger.info("Requirements installed successfully!")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error installing requirements: {e}")
            self.clean_up()
            sys.exit(1)

    def _build_request(self):
        """Create a mock inference request for testing the model."""

        model_version_proto = self.builder.get_model_version_proto()
        model_version_proto.id = "model_version"

        return service_pb2.PostModelOutputsRequest(
            model=resources_pb2.Model(model_version=model_version_proto),
            inputs=[
                resources_pb2.Input(
                    data=resources_pb2.Data(
                        text=resources_pb2.Text(raw="How many people live in new york?"),
                        image=resources_pb2.Image(
                            url="https://samples.clarifai.com/metro-north.jpg"
                        ),
                        audio=resources_pb2.Audio(
                            url="https://samples.clarifai.com/GoodMorning.wav"
                        ),
                        video=resources_pb2.Video(url="https://samples.clarifai.com/beer.mp4"),
                    )
                )
            ],
        )

    def _run_test(self):
        """Test the model locally by making a prediction."""
        # Create the model
        model = self.builder.create_model_instance()
        # call its test method, if it has one
        if hasattr(model, "test"):
            try:
                model.test()
                logger.info("Model tested successfully!")
            except Exception as e:
                logger.error(f"Error occurred while testing the model: {e}")
                traceback.print_exc()

    def test_model(self):
        """Test the model by running it locally in the virtual environment."""

        import_path = repr(os.path.dirname(os.path.abspath(__file__)))
        model_path = repr(self.model_path)

        command_string = (
            f"import sys; "
            f"sys.path.append({import_path}); "
            f"from model_run_locally import ModelRunLocally; "
            f"ModelRunLocally({model_path})._run_test()"
        )

        command = [self.python_executable, "-c", command_string]
        process = None
        try:
            logger.info("Testing the model locally...")
            process = subprocess.Popen(command)
            # Wait for the process to complete
            process.wait()
            if process.returncode == 0:
                logger.info("Model tested successfully!")
            if process.returncode != 0:
                raise subprocess.CalledProcessError(process.returncode, command)
        except subprocess.CalledProcessError as e:
            logger.error(f"Error testing the model: {e}")
            sys.exit(1)
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            sys.exit(1)
        finally:
            # After the function runs, check if the process is still running
            if process and process.poll() is None:
                logger.info("Process is still running. Terminating process.")
                process.terminate()
                try:
                    process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    logger.info("Process did not terminate gracefully. Killing process.")
                    # Kill the process if it doesn't terminate after 5 seconds
                    process.kill()

    # run the model server
    def run_model_server(self, port=8080):
        """Run the Clarifai Runners's model server."""

        command = [
            self.python_executable,
            "-m",
            "clarifai.runners.server",
            "--model_path",
            self.model_path,
            "--grpc",
            "--port",
            str(port),
        ]
        try:
            logger.info(
                f"Starting model server at localhost:{port} with the model at {self.model_path}..."
            )
            subprocess.check_call(command)
            logger.info("Model server started successfully and running at localhost:{port}")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error running model server: {e}")
            self.clean_up()
            sys.exit(1)

    def _docker_hash(self):
        """Generate a hash of the combined requirements file and Dockefile"""
        with open(self.requirements_file, "r") as f:
            requirements_hash = hashlib.md5(f.read().encode('utf-8')).hexdigest()
        with open(os.path.join(self.model_path, "Dockerfile"), "r") as f:
            dockerfile_hash = hashlib.md5(f.read().encode('utf-8')).hexdigest()

        return hashlib.md5(f"{requirements_hash}{dockerfile_hash}".encode('utf-8')).hexdigest()

    def is_docker_installed(self):
        """Checks if Docker is installed on the system."""
        try:
            logger.info("Checking if Docker is installed...")
            subprocess.run(["docker", "--version"], check=True)
            return True
        except subprocess.CalledProcessError:
            logger.error(
                "Docker is not installed! Please install Docker to run the model in a container."
            )
            return False

    def build_docker_image(
        self,
        image_name="model_image",
    ):
        """Build the docker image using the Dockerfile in the model directory."""
        try:
            logger.info(f"Building docker image from Dockerfile in {self.model_path}...")

            # since we don't want to copy the model directory into the container, we need to modify the Dockerfile and comment out the COPY instruction
            dockerfile_path = os.path.join(self.model_path, "Dockerfile")
            # Read the Dockerfile
            with open(dockerfile_path, 'r') as file:
                lines = file.readlines()

            # Comment out the COPY instruction that copies the current folder
            modified_lines = []
            for line in lines:
                if 'COPY' in line and '/home/nonroot/main' in line:
                    modified_lines.append(f'# {line}')
                elif 'download-checkpoints' in line and '/home/nonroot/main' in line:
                    modified_lines.append(f'# {line}')
                else:
                    modified_lines.append(line)

            # Create a temporary directory to store the modified Dockerfile
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_dockerfile_path = os.path.join(temp_dir, "Dockerfile.temp")

                # Write the modified Dockerfile to the temporary file
                with open(temp_dockerfile_path, 'w') as file:
                    file.writelines(modified_lines)

                # Build the Docker image using the temporary Dockerfile
                subprocess.check_call(
                    [
                        'docker',
                        'build',
                        '-t',
                        image_name,
                        '-f',
                        temp_dockerfile_path,
                        self.model_path,
                    ]
                )
            logger.info(f"Docker image '{image_name}' built successfully!")
        except subprocess.CalledProcessError as e:
            logger.info(f"Error occurred while building the Docker image: {e}")
            sys.exit(1)

    def docker_image_exists(self, image_name):
        """Check if the Docker image exists."""
        try:
            logger.info(f"Checking if Docker image '{image_name}' exists...")
            subprocess.run(["docker", "inspect", image_name], check=True)
            logger.info(f"Docker image '{image_name}' exists!")
            return True
        except subprocess.CalledProcessError:
            logger.info(f"Docker image '{image_name}' does not exist!")
            return False

    def _gpu_is_available(self):
        """
        Checks if nvidia-smi is available, indicating a GPU is likely accessible.
        """
        return shutil.which("nvidia-smi") is not None

    def run_docker_container(
        self, image_name, container_name="clarifai-model-container", port=8080, env_vars=None
    ):
        """Runs a Docker container from the specified image."""
        try:
            logger.info(
                f"Running Docker container '{container_name}' from image '{image_name}'..."
            )
            # Base docker run command
            cmd = ["docker", "run", "--name", container_name, '--rm', "--network", "host"]
            if self._gpu_is_available():
                cmd.extend(["--gpus", "all"])
            # Add volume mappings
            cmd.extend(["-v", f"{self.model_path}:/home/nonroot/main"])
            # Add environment variables
            if env_vars:
                for key, value in env_vars.items():
                    cmd.extend(["-e", f"{key}={value}"])
            # Add the image name
            cmd.append(image_name)
            # update the CMD to run the server
            cmd.extend(["--model_path", "/home/nonroot/main", "--grpc", "--port", str(port)])
            # Run the container
            process = subprocess.Popen(
                cmd,
            )
            logger.info(
                f"Docker container '{container_name}' is running successfully! access the model at http://localhost:{port}"
            )

            # Function to handle Ctrl+C (SIGINT) gracefully
            def signal_handler(sig, frame):
                logger.info(f"Stopping Docker container '{container_name}'...")
                subprocess.run(["docker", "stop", container_name], check=True)
                process.terminate()
                logger.info(f"Docker container '{container_name}' stopped successfully!")
                time.sleep(1)
                sys.exit(0)

            # Register the signal handler for SIGINT (Ctrl+C)
            signal.signal(signal.SIGINT, signal_handler)
            # Wait for the process to finish (keeps the container running until it's stopped)
            process.wait()
        except subprocess.CalledProcessError as e:
            logger.info(f"Error occurred while running the Docker container: {e}")
            sys.exit(1)
        except Exception as e:
            logger.info(f"Error occurred while running the Docker container: {e}")
            sys.exit(1)

    def test_model_container(
        self, image_name, container_name="clarifai-model-container", env_vars=None
    ):
        """Test the model inside the Docker container."""
        try:
            logger.info("Testing the model inside the Docker container...")
            # Base docker run command
            cmd = ["docker", "run", "--name", container_name, '--rm', "--network", "host"]
            if self._gpu_is_available():
                cmd.extend(["--gpus", "all"])
            # update the entrypoint for testing the model
            cmd.extend(["--entrypoint", "python"])
            # Add volume mappings
            cmd.extend(["-v", f"{self.model_path}:/home/nonroot/main"])
            # Add environment variables
            if env_vars:
                for key, value in env_vars.items():
                    cmd.extend(["-e", f"{key}={value}"])
            # Add the image name
            cmd.append(image_name)
            # update the CMD to test the model inside the container
            cmd.extend(
                [
                    "-c",
                    "from clarifai.runners.models.model_run_locally import ModelRunLocally; ModelRunLocally('/home/nonroot/main')._run_test()",
                ]
            )
            # Run the container
            subprocess.check_call(cmd)
            logger.info("Model tested successfully!")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error testing the model inside the Docker container: {e}")
            sys.exit(1)

    def container_exists(self, container_name="clarifai-model-container"):
        """Check if the Docker container exists."""
        try:
            # Run docker ps -a to list all containers (running and stopped)
            result = subprocess.run(
                [
                    "docker",
                    "ps",
                    "-a",
                    "--filter",
                    f"name={container_name}",
                    "--format",
                    "{{.Names}}",
                ],
                check=True,
                capture_output=True,
                text=True,
            )
            # If the container name is returned, it exists
            if result.stdout.strip() == container_name:
                logger.info(f"Docker container '{container_name}' exists.")
                return True
            else:
                return False
        except subprocess.CalledProcessError as e:
            logger.error(f"Error occurred while checking if container exists: {e}")
            return False

    def stop_docker_container(self, container_name="clarifai-model-container"):
        """Stop the Docker container if it's running."""
        try:
            # Check if the container is running
            result = subprocess.run(
                ["docker", "ps", "--filter", f"name={container_name}", "--format", "{{.Names}}"],
                check=True,
                capture_output=True,
                text=True,
            )
            if result.stdout.strip() == container_name:
                logger.info(f"Docker container '{container_name}' is running. Stopping it...")
                subprocess.run(["docker", "stop", container_name], check=True)
                logger.info(f"Docker container '{container_name}' stopped successfully!")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error occurred while stopping the Docker container: {e}")

    def remove_docker_container(self, container_name="clarifai-model-container"):
        """Remove the Docker container."""
        try:
            logger.info(f"Removing Docker container '{container_name}'...")
            subprocess.run(["docker", "rm", container_name], check=True)
            logger.info(f"Docker container '{container_name}' removed successfully!")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error occurred while removing the Docker container: {e}")

    def remove_docker_image(self, image_name):
        """Remove the Docker image."""
        try:
            logger.info(f"Removing Docker image '{image_name}'...")
            subprocess.run(["docker", "rmi", image_name], check=True)
            logger.info(f"Docker image '{image_name}' removed successfully!")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error occurred while removing the Docker image: {e}")

    def clean_up(self):
        """Clean up the temporary virtual environment."""
        if os.path.exists(self.temp_dir):
            logger.info("Cleaning up temporary virtual environment...")
            shutil.rmtree(self.temp_dir)


def main(
    model_path,
    run_model_server=False,
    inside_container=False,
    port=8080,
    keep_env=False,
    keep_image=False,
    skip_dockerfile: bool = False,
):
    manager = ModelRunLocally(model_path)
    # get whatever stage is in config.yaml to force download now
    # also always write to where upload/build wants to, not the /tmp folder that runtime stage uses
    if inside_container:
        if not manager.is_docker_installed():
            sys.exit(1)
        if not skip_dockerfile:
            manager.builder.create_dockerfile()
        image_tag = manager._docker_hash()
        model_id = manager.config['model']['id'].lower()
        # must be in lowercase
        image_name = f"{model_id}:{image_tag}"
        container_name = model_id
        if not manager.docker_image_exists(image_name):
            manager.build_docker_image(image_name=image_name)
        try:
            if run_model_server:
                manager.run_docker_container(
                    image_name=image_name, container_name=container_name, port=port
                )
            else:
                manager.test_model_container(image_name=image_name, container_name=container_name)
        finally:
            if manager.container_exists(container_name):
                manager.stop_docker_container(container_name)
                manager.remove_docker_container(container_name=container_name)
            if not keep_image:
                manager.remove_docker_image(image_name)

    else:
        try:
            use_existing_env = manager.create_temp_venv()
            if not use_existing_env:
                manager.install_requirements()
            if run_model_server:
                manager.run_model_server(port)
            else:
                manager.test_model()
        finally:
            if not keep_env:
                manager.clean_up()



from typing import List, Tuple

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2 as respb2


def parse_eval_annotation_classifier(
    eval_metrics: respb2.EvalMetrics,
) -> Tuple[np.array, np.array, list, List[respb2.Input]]:
    test_set = eval_metrics.test_set
    # get concept ids
    concept_ids = [each.id for each in test_set[0].predicted_concepts]
    concept_ids.sort()
    # get test set
    y_preds = []
    y = []
    inputs = []
    for data in test_set:

        def _to_array(_data):
            cps = [0] * len(concept_ids)
            for each in _data:
                cps[concept_ids.index(each.id)] = each.value
            return np.asarray(cps)

        y_preds.append(_to_array(data.predicted_concepts))
        y.append(_to_array(data.ground_truth_concepts))
        inputs.append(data.input)

    return np.asarray(y), np.asarray(y_preds), concept_ids, inputs


def parse_eval_annotation_detector(
    eval_metrics: respb2.EvalMetrics, normalized_box: bool = False, box_style: str = "xyxy"
) -> Tuple[np.array, np.array, list, List[respb2.Input]]:
    BOX_STYLES = ["xyxy", "xywh"]
    assert box_style in BOX_STYLES, ValueError(f"Expected box_style in {BOX_STYLES}")

    concept_ids = list(set([each.concept.id for each in eval_metrics.metrics_by_class]))
    concept_ids.sort()

    def _get_box_annot(field, img_height, img_width):
        xyxy_concept_score = []
        for each in field:
            box = each.region_info.bounding_box
            x1 = box.left_col * img_width
            y1 = box.top_row * img_height
            x2 = box.right_col * img_width
            y2 = box.bottom_row * img_height
            score = each.data.concepts[0].value
            concept = each.data.concepts[0].id
            concept_index = concept_ids.index(concept)
            if box_style == "xyxy":
                xyxy_concept_score.append([x1, y1, x2, y2, concept_index, score])
            else:
                w = abs(x1 - x2)
                h = abs(y1 - y2)
                xyxy_concept_score.append([x1, y1, w, h, concept_index, score])

        return np.asarray(xyxy_concept_score)

    inputs = []
    pred_xyxy_concept_score = []
    gt_xyxy_concept_score = []
    for input_data in eval_metrics.test_set:
        _input = input_data.input
        img_height = _input.data.image.image_info.height if not normalized_box else 1.0
        img_width = _input.data.image.image_info.height if not normalized_box else 1.0
        _pred_xyxy_concept_score = _get_box_annot(
            input_data.predicted_annotation.data.regions,
            img_height=img_height,
            img_width=img_width,
        )
        _gt_xyxy_concept_score = _get_box_annot(
            input_data.ground_truth_annotation.data.regions,
            img_height=img_height,
            img_width=img_width,
        )

        pred_xyxy_concept_score.append(_pred_xyxy_concept_score)
        gt_xyxy_concept_score.append(_gt_xyxy_concept_score)
        inputs.append(_input)

    return (
        np.asarray(gt_xyxy_concept_score),
        np.asarray(pred_xyxy_concept_score),
        concept_ids,
        inputs,
    )


def parse_eval_annotation_detector_coco(
    eval_metrics: respb2.EvalMetrics,
) -> Tuple[np.array, np.array, list, List[respb2.Input]]:
    gts, preds, concept_ids, inputs = parse_eval_annotation_detector(
        eval_metrics=eval_metrics, normalized_box=False, box_style="xywh"
    )

    def _make_box_annot(data, input_data, accum_id, is_pred=True):
        img_id = input_data.id
        img_url = input_data.data.image.url
        img_height = input_data.data.image.image_info.height
        img_width = input_data.data.image.image_info.height
        image = {
            "id": img_id,
            "file_name": img_url,
            "width": img_width,
            "height": img_height,
        }
        annotations = []
        for i, d in enumerate(data.tolist()):
            area = d[2] * d[3]
            box = {
                "iscrowd": 0,
                "ignore": 0,
                "image_id": img_id,
                "bbox": d[:4],
                "area": area,
                "segmentation": [],
                "category_id": d[4],
                "id": accum_id + i,
            }
            if is_pred:
                box["score"] = d[5]
            annotations.append(box)

        return image, annotations

    categories = [
        {"supercategory": "none", "id": i, "name": label} for i, label in enumerate(concept_ids)
    ]

    accum_pred_ids = 0
    accum_gt_ids = 0
    pred_images = []
    pred_boxes = []
    gt_images = []
    gt_boxes = []
    for ith, input_proto in enumerate(inputs):
        pred = preds[ith]
        gt = gts[ith]
        pred_img, pred_box = _make_box_annot(
            pred, accum_id=accum_pred_ids, input_data=input_proto, is_pred=True
        )
        gt_img, gt_box = _make_box_annot(
            gt, accum_id=accum_gt_ids, input_data=input_proto, is_pred=False
        )

        accum_pred_ids += len(pred)
        pred_images.append(pred_img)
        pred_boxes += pred_box

        accum_gt_ids += len(gt)
        gt_images.append(gt_img)
        gt_boxes += gt_box

    pred_annots = {"images": pred_images, "annotations": pred_boxes, "categories": categories}
    gt_annots = {"images": gt_images, "annotations": gt_boxes, "categories": categories}

    return gt_annots, pred_annots



from math import ceil
from typing import Any, Callable, Dict, Generator

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from schema import SchemaError

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.search import (
    DEFAULT_SEARCH_ALGORITHM,
    DEFAULT_SEARCH_METRIC,
    DEFAULT_TOP_K,
)
from clarifai.errors import UserError
from clarifai.schema.search import get_schema
from clarifai.utils.constants import DEFAULT_BASE


class Search(Lister, BaseClient):
    def __init__(
        self,
        user_id: str,
        app_id: str,
        top_k: int = None,
        metric: str = DEFAULT_SEARCH_METRIC,
        algorithm: str = DEFAULT_SEARCH_ALGORITHM,
        pagination: bool = False,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
    ):
        """Initialize the Search object.

        Args:
            user_id (str): User ID.
            app_id (str): App ID.
            top_k (int, optional): Top K results to retrieve. Defaults to 10.
            metric (str, optional): Similarity metric (either 'cosine' or 'euclidean'). Defaults to 'cosine'.
            alogrithm (str, optional): Search algorithm (either 'nearest_neighbor' or 'brute_force'). Defaults to 'nearest_neighbor'.
            pagination (bool, optional): Enable pagination. Defaults to False.
            base_url (str, optional): Base API url. Defaults to "https://api.clarifai.com".
            pat (str, optional): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.

        Raises:
            UserError: If the metric is not 'cosine' or 'euclidean'.
            UserError: If the algorithm is not 'nearest_neighbor' or 'brute_force'.
        """
        if metric not in ["cosine", "euclidean"]:
            raise UserError("Metric should be either cosine or euclidean")
        if algorithm not in ["nearest_neighbor", "brute_force"]:
            raise UserError("Algorithm should be either nearest_neighbor or brute_force")
        if metric == "cosine" and algorithm == "nearest_neighbor":
            raise UserError(
                "Cosine distance metric is not supported with nearest neighbor algorithm"
            )
        if top_k and pagination:
            raise UserError(
                "top_k and pagination cannot be used together. Please set pagination to False."
            )
        if not top_k and not pagination:
            top_k = DEFAULT_TOP_K

        self.user_id = user_id
        self.app_id = app_id
        self.metric_distance = dict(cosine="COSINE_DISTANCE", euclidean="EUCLIDEAN_DISTANCE")[
            metric
        ]
        self.algorithm = algorithm
        self.data_proto = resources_pb2.Data()
        self.top_k = top_k
        self.pagination = pagination
        self.inputs = Inputs(
            user_id=self.user_id, app_id=self.app_id, pat=pat, token=token, base_url=base_url
        )
        self.rank_filter_schema = get_schema()
        BaseClient.__init__(
            self,
            user_id=self.user_id,
            app_id=self.app_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self, page_size=1000)

    def _get_annot_proto(self, **kwargs) -> resources_pb2.Annotation:
        """Get an Annotation proto message based on keyword arguments.

        Args:
            **kwargs: Keyword arguments specifying the annotation data.
                     Supported keys:
                     - image_bytes (bytes): Raw image bytes
                     - image_url (str): URL to an image
                     - concepts (List[Dict]): List of concept dictionaries
                     - metadata (Dict): Metadata dictionary
                     - geo_longitude (float): Geographic longitude
                     - geo_latitude (float): Geographic latitude

        Returns:
            resources_pb2.Annotation: An Annotation proto message with the specified data.
        """
        if not kwargs:
            return resources_pb2.Annotation()

        self.data_proto = resources_pb2.Data()
        for key, value in kwargs.items():
            if key == "image_bytes":
                image_proto = self.inputs.get_input_from_bytes("", image_bytes=value).data.image
                self.data_proto.image.CopyFrom(image_proto)

            elif key == "image_url":
                image_proto = self.inputs.get_input_from_url("", image_url=value).data.image
                self.data_proto.image.CopyFrom(image_proto)

            elif key == "concepts":
                for concept in value:
                    concept_proto = resources_pb2.Concept(**concept)
                    self.data_proto.concepts.add().CopyFrom(concept_proto)

            elif key == "text_raw":
                text_proto = self.inputs.get_input_from_bytes(
                    "", text_bytes=bytes(value, 'utf-8')
                ).data.text
                self.data_proto.text.CopyFrom(text_proto)

            elif key == "metadata":
                metadata_struct = Struct()
                metadata_struct.update(value)
                self.data_proto.metadata.CopyFrom(metadata_struct)

            elif key == "geo_point":
                geo_point_proto = self._get_geo_point_proto(
                    value["longitude"], value["latitude"], value["geo_limit"]
                )
                self.data_proto.geo.CopyFrom(geo_point_proto)

            else:
                raise UserError(f"kwargs contain key that is not supported: {key}")
        return resources_pb2.Annotation(data=self.data_proto)

    def _get_input_proto(self, **kwargs) -> resources_pb2.Input:
        """Get an Input proto message based on keyword arguments.

        Args:
            **kwargs: Keyword arguments specifying the input data.
                     Supported keys:
                     - input_types (List[str]): List of input types ('image', 'text', 'audio', 'video')
                     - dataset_ids (List[str]): List of dataset IDs to filter by
                     - image_bytes (bytes): Raw image bytes
                     - image_url (str): URL to an image
                     - text_raw (str): Raw text content
                     - concepts (List[Dict]): List of concept dictionaries
                     - metadata (Dict): Metadata dictionary
                     - geo_longitude (float): Geographic longitude
                     - geo_latitude (float): Geographic latitude

        Returns:
            resources_pb2.Input: An Input proto message with the specified data.
        """
        if not kwargs:
            return resources_pb2.Input()

        self.input_proto = resources_pb2.Input()
        self.data_proto = resources_pb2.Data()
        for key, value in kwargs.items():
            if key == "input_types":
                for input_type in value:
                    if input_type == "image":
                        self.data_proto.image.CopyFrom(resources_pb2.Image())
                    elif input_type == "text":
                        self.data_proto.text.CopyFrom(resources_pb2.Text())
                    elif input_type == "audio":
                        self.data_proto.audio.CopyFrom(resources_pb2.Audio())
                    elif input_type == "video":
                        self.data_proto.video.CopyFrom(resources_pb2.Video())
                self.input_proto.data.CopyFrom(self.data_proto)
            elif key == "input_dataset_ids":
                self.input_proto.dataset_ids.extend(value)
            elif key == "input_status_code":
                self.input_proto.status.code = value
            else:
                raise UserError(f"kwargs contain key that is not supported: {key}")
        return self.input_proto

    def _get_geo_point_proto(
        self, longitude: float, latitude: float, geo_limit: float
    ) -> resources_pb2.Geo:
        """Get a GeoPoint proto message based on geographical data.

        Args:
            longitude (float): Longitude coordinate.
            latitude (float): Latitude coordinate.
            geo_limit (float): Geographical limit.

        Returns:
            resources_pb2.Geo: A Geo proto message.
        """
        return resources_pb2.Geo(
            geo_point=resources_pb2.GeoPoint(longitude=longitude, latitude=latitude),
            geo_limit=resources_pb2.GeoLimit(type="withinKilometers", value=geo_limit),
        )

    def _list_topk_generator(
        self, endpoint: Callable[..., Any], proto_message: Any, request_data: Dict[str, Any]
    ) -> Generator[Dict[str, Any], None, None]:
        """Lists top-k results with pagination support.

        This method handles pagination for search results when top_k is specified,
        automatically calculating the required number of pages and per-page limits.

        Args:
            endpoint (Callable[..., Any]): The gRPC endpoint method to call for search.
            proto_message (Any): The protobuf message class for the request.
            request_data (Dict[str, Any]): The base request data dictionary.

        Yields:
            Dict[str, Any]: Individual search result items from the API response.

        Raises:
            UserError: If pagination limits are exceeded or top_k is too large.
            Exception: If the API request fails.
        """
        max_pages = ceil(self.top_k / self.default_page_size)
        total_hits = 0
        page = 1
        while page <= max_pages:
            if page == max_pages:
                per_page = self.top_k - total_hits
            else:
                per_page = self.default_page_size
            request_data['pagination'] = service_pb2.Pagination(page=page, per_page=per_page)
            response = self._grpc_request(endpoint, proto_message(**request_data))
            dict_response = MessageToDict(response, preserving_proto_field_name=True)
            if response.status.code != status_code_pb2.SUCCESS:
                if "page * perPage cannot exceed" in str(response.status.details):
                    msg = (
                        f"Your top_k is set to {self.top_k}. "
                        f"The current pagination settings exceed the limit. Please reach out to "
                        f"support@clarifai.com to request an increase for your use case.\n"
                        f"req_id: {response.status.req_id}"
                    )
                    raise UserError(msg)
                else:
                    raise Exception(f"Listing failed with response {response!r}")

            if 'hits' not in list(dict_response.keys()):
                break
            page += 1
            total_hits += per_page
            yield response

    def _list_all_pages_generator(
        self,
        endpoint: Callable,
        proto_message: Any,
        request_data: Dict[str, Any],
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """Lists pages of a resource.

        Args:
            endpoint (Callable): The endpoint to call.
            proto_message (Any): The proto message to use.
            request_data (dict): The request data to use.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            response_dict: The next item in the listing.
        """
        page = 1 if not page_no else page_no
        if page_no and not per_page:
            per_page = self.default_page_size
        while True:
            request_data['pagination'] = service_pb2.Pagination(page=page, per_page=per_page)
            response = self._grpc_request(endpoint, proto_message(**request_data))
            dict_response = MessageToDict(response, preserving_proto_field_name=True)
            if response.status.code != status_code_pb2.SUCCESS:
                if "page * perPage cannot exceed" in str(response.status.details):
                    msg = (
                        f"Your pagination is set to {page_no * per_page}. "
                        f"The current pagination settings exceed the limit. Please reach out to "
                        f"support@clarifai.com to request an increase for your use case.\n"
                        f"req_id: {response.status.req_id}"
                    )
                    raise UserError(msg)
                else:
                    raise Exception(f"Listing failed with response {response!r}")
            if 'hits' not in list(dict_response.keys()):
                break
            yield response
            if page_no is not None or per_page is not None:
                break
            page += 1

    def query(
        self,
        ranks=[{}],
        filters=[{}],
        page_no: int = None,
        per_page: int = None,
    ):
        """Perform a query with rank and filters.

        Args:
            ranks (List[Dict], optional): List of rank parameters. Defaults to [{}].
            filters (List[Dict], optional): List of filter parameters. Defaults to [{}].
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Returns:
            Generator[Dict[str, Any], None, None]: A generator of query results.

        Examples:
            Get successful inputs of type image or text
            >>> from clarifai.client.search import Search
            >>> search = Search(user_id='user_id', app_id='app_id', top_k=10, metric='cosine')
            # This performs OR operation on input_types and input_status_code
            >>> res = search.query(filters=[{'input_types': ['image', 'text'], 'input_status_code': 30000}])

            Vector search over inputs
            >>> from clarifai.client.search import Search
            >>> search = Search(user_id='user_id', app_id='app_id' , metric='cosine', pagination=True)
            >>> res = search.query(ranks=[{'image_url': 'https://samples.clarifai.com/dog.tiff'}],page_no=2, per_page=5)

        Note:
            For schema of rank and filter, please refer to [schema](https://github.com/Clarifai/clarifai-python/tree/master/clarifai/schema/search.py).
            For more detailed search examples, please refer to [examples](https://github.com/Clarifai/examples/tree/main/search).
        """
        if not self.pagination and (per_page or page_no):
            raise UserError(
                "Pagination settings are only available when pagination is enabled."
                "Please set Search(pagination=True) while initializing Search()."
            )
        try:
            self.rank_filter_schema.validate(ranks)
            self.rank_filter_schema.validate(filters)
        except SchemaError as err:
            raise UserError(f"Invalid rank or filter input: {err}")

        # For each rank, create a Rank proto message
        # For ranks it only allows resources_pb2.Annotation proto, so no need of splitting protos into annotation and input.
        all_ranks = []
        for rank_dict in ranks:
            rank_annot_proto = self._get_annot_proto(**rank_dict)
            all_ranks.append(resources_pb2.Rank(annotation=rank_annot_proto))

        all_filters = []
        # check for filters which is compatible with input proto
        for each_filter in filters:
            input_dict = {
                key: each_filter.pop(key)
                for key in ['input_types', 'input_dataset_ids', 'input_status_code']
                if key in each_filter
            }

            all_filters.append(
                resources_pb2.Filter(
                    annotation=self._get_annot_proto(**each_filter),
                    input=self._get_input_proto(**input_dict),
                )
            )

        # Create a PostInputsSearchesRequest proto message
        request_data = dict(
            user_app_id=self.user_app_id,
            searches=[
                resources_pb2.Search(
                    query=resources_pb2.Query(ranks=all_ranks, filters=all_filters),
                    algorithm=self.algorithm,
                    metric=self.metric_distance,
                )
            ],
        )
        # Calls PostInputsSearches for annotation ranks, input filters
        if self.pagination:
            return self._list_all_pages_generator(
                self.STUB.PostInputsSearches,
                service_pb2.PostInputsSearchesRequest,
                request_data,
                page_no,
                per_page,
            )
        return self._list_topk_generator(
            self.STUB.PostInputsSearches, service_pb2.PostInputsSearchesRequest, request_data
        )



import os
from itertools import tee
from typing import Iterator

from clarifai_grpc.grpc.api import service_pb2, service_pb2_grpc
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.utils.secrets import inject_secrets

from ..utils.url_fetcher import ensure_urls_downloaded

_RAISE_EXCEPTIONS = os.getenv("RAISE_EXCEPTIONS", "false").lower() in ("true", "1")


class ModelServicer(service_pb2_grpc.V2Servicer):
    """
    This is the servicer that will handle the gRPC requests from either the dev server or runner loop.
    """

    def __init__(self, model):
        """
        Args:
            model: The class that will handle the model logic. Must implement predict(),
        generate(), stream().
        """
        self.model = model

        # Try to create auth helper from environment variables if available
        self._auth_helper = None
        try:
            user_id = os.environ.get("CLARIFAI_USER_ID")
            pat = os.environ.get("CLARIFAI_PAT")
            token = os.environ.get("CLARIFAI_SESSION_TOKEN")
            base_url = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")

            if user_id and (pat or token):
                self._auth_helper = ClarifaiAuthHelper(
                    user_id=user_id,
                    app_id="",  # app_id not needed for URL fetching
                    pat=pat or "",
                    token=token or "",
                    base=base_url,
                    validate=False,  # Don't validate since app_id is empty
                )
        except Exception:
            # If auth helper creation fails, proceed without authentication
            self._auth_helper = None

    def PostModelOutputs(
        self, request: service_pb2.PostModelOutputsRequest, context=None
    ) -> service_pb2.MultiOutputResponse:
        """
        This is the method that will be called when the servicer is run. It takes in an input and
        returns an output.
        """

        # Download any urls that are not already bytes.
        ensure_urls_downloaded(request, auth_helper=self._auth_helper)
        inject_secrets(request)

        try:
            return self.model.predict_wrapper(request)
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            return service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.MODEL_PREDICTION_FAILED,
                    description="Failed",
                    details="",
                    internal_details=str(e),
                )
            )

    def GenerateModelOutputs(
        self, request: service_pb2.PostModelOutputsRequest, context=None
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        """
        This is the method that will be called when the servicer is run. It takes in an input and
        returns an output.
        """
        # Download any urls that are not already bytes.
        ensure_urls_downloaded(request, auth_helper=self._auth_helper)
        inject_secrets(request)

        try:
            yield from self.model.generate_wrapper(request)
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.MODEL_PREDICTION_FAILED,
                    description="Failed",
                    details="",
                    internal_details=str(e),
                )
            )

    def StreamModelOutputs(
        self, request: Iterator[service_pb2.PostModelOutputsRequest], context=None
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        """
        This is the method that will be called when the servicer is run. It takes in an input and
        returns an output.
        """
        # Duplicate the iterator
        request, request_copy = tee(request)

        # Download any urls that are not already bytes.
        for req in request:
            ensure_urls_downloaded(req, auth_helper=self._auth_helper)
            inject_secrets(req)

        try:
            yield from self.model.stream_wrapper(request_copy)
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.MODEL_PREDICTION_FAILED,
                    description="Failed",
                    details="",
                    internal_details=str(e),
                )
            )

    def set_model(self, model):
        self.model = model



import os

registry = os.environ.get('CLARIFAI_BASE_IMAGE_REGISTRY', 'public.ecr.aws/clarifai-models')

GIT_SHA = "42938da8e33b0f37ee7db16b83631da94c2348b9"

AMD_GIT_SHA = "81e942130173f54927e7c9a65aabc7e32780616d"

PYTHON_BASE_IMAGE = registry + '/python-base:{python_version}-' + GIT_SHA
TORCH_BASE_IMAGE = registry + '/torch:{torch_version}-py{python_version}-{gpu_version}-' + GIT_SHA

AMD_PYTHON_BASE_IMAGE = registry + '/amd-python-base:{python_version}-' + AMD_GIT_SHA
AMD_TORCH_BASE_IMAGE = (
    registry + '/amd-torch:{torch_version}-py{python_version}-{gpu_version}-' + AMD_GIT_SHA
)
AMD_VLLM_BASE_IMAGE = (
    registry + '/amd-vllm:{torch_version}-py{python_version}-{gpu_version}-' + AMD_GIT_SHA
)

# List of available python base images
AVAILABLE_PYTHON_IMAGES = ['3.11', '3.12']

DEFAULT_PYTHON_VERSION = '3.12'

DEFAULT_AMD_TORCH_VERSION = '2.8.0.dev20250514'

DEFAULT_AMD_GPU_VERSION = 'rocm6.4'

# By default we download at runtime.
DEFAULT_DOWNLOAD_CHECKPOINT_WHEN = "runtime"

# Folder for downloading checkpoints at runtime.
DEFAULT_RUNTIME_DOWNLOAD_PATH = os.path.join(os.sep, "tmp", ".cache")

# List of available torch images
# Keep sorted by most recent cuda version.
AVAILABLE_TORCH_IMAGES = [
    '2.4.1-py3.11-cu124',
    '2.5.1-py3.11-cu124',
    '2.4.1-py3.12-cu124',
    '2.5.1-py3.12-cu124',
    '2.6.0-py3.12-cu126',
    '2.7.0-py3.12-cu128',
    '2.7.0-py3.12-rocm6.3',
]

CONCEPTS_REQUIRED_MODEL_TYPE = [
    'visual-classifier',
    'visual-detector',
    'visual-segmenter',
    'text-classifier',
]



import fnmatch
import importlib.util
import json
import os
import shutil

import requests

from clarifai.runners.utils.const import CONCEPTS_REQUIRED_MODEL_TYPE
from clarifai.utils.logging import logger


class HuggingFaceLoader:
    HF_DOWNLOAD_TEXT = "The 'huggingface_hub' package is not installed. Please install it using 'pip install huggingface_hub'."

    def __init__(self, repo_id=None, token=None, model_type_id=None):
        self.repo_id = repo_id
        self.token = token
        self.clarifai_model_type_id = model_type_id
        if token:
            if self.validate_hftoken(token):
                try:
                    from huggingface_hub import login
                except ImportError:
                    raise ImportError(self.HF_DOWNLOAD_TEXT)
                login(token=token)
                logger.info("Hugging Face token validated")
            else:
                self.token = None
                logger.info("Continuing without Hugging Face token")

    @classmethod
    def validate_hftoken(cls, hf_token: str):
        try:
            if importlib.util.find_spec("huggingface_hub") is None:
                raise ImportError(cls.HF_DOWNLOAD_TEXT)
            from huggingface_hub import HfApi

            api = HfApi()
            api.whoami(token=hf_token)
            return True
        except Exception as e:
            logger.error(
                f"Invalid Hugging Face token provided in the config file, this might cause issues with downloading the restricted model checkpoints. Failed reason: {e}"
            )
            return False

    def download_checkpoints(
        self, checkpoint_path: str, allowed_file_patterns=None, ignore_file_patterns=None
    ):
        # throw error if huggingface_hub wasn't installed
        try:
            from huggingface_hub import snapshot_download
        except ImportError:
            raise ImportError(self.HF_DOWNLOAD_TEXT)
        if os.path.exists(checkpoint_path) and self.validate_download(
            checkpoint_path, allowed_file_patterns, ignore_file_patterns
        ):
            logger.info("Checkpoints already exist")
            return True
        else:
            os.makedirs(checkpoint_path, exist_ok=True)
            try:
                is_hf_model_exists = self.validate_hf_model()
                if not is_hf_model_exists:
                    return False

                self.ignore_patterns = self._get_ignore_patterns()
                if ignore_file_patterns:
                    if self.ignore_patterns:
                        self.ignore_patterns.extend(ignore_file_patterns)
                    else:
                        self.ignore_patterns = ignore_file_patterns

                repo_files_to_download = self.get_repo_files_list(
                    allowed_file_patterns=allowed_file_patterns,
                    ignore_file_patterns=self.ignore_patterns,
                )
                total_size = self.get_huggingface_checkpoint_total_size(
                    self.repo_id, checkpoint_files_list=repo_files_to_download
                )
                total_size = total_size / (1024**2)
                logger.info(f"Total download size: {total_size:.2f} MB")

                logger.info("Downloading model checkpoints...")
                snapshot_download(
                    repo_id=self.repo_id,
                    local_dir=checkpoint_path,
                    local_dir_use_symlinks=False,
                    allow_patterns=allowed_file_patterns,
                    ignore_patterns=self.ignore_patterns,
                )
                # Remove the `.cache` folder if it exists
                cache_path = os.path.join(checkpoint_path, ".cache")
                if os.path.exists(cache_path) and os.path.isdir(cache_path):
                    shutil.rmtree(cache_path)

            except Exception as e:
                logger.error(f"Error downloading model checkpoints {e}")
                return False
            finally:
                is_downloaded = self.validate_download(
                    checkpoint_path, allowed_file_patterns, ignore_file_patterns
                )
                if not is_downloaded:
                    logger.error("Error validating downloaded model checkpoints")
                    return False
            return True

    def download_config(self, checkpoint_path: str):
        # throw error if huggingface_hub wasn't installed
        try:
            from huggingface_hub import hf_hub_download
        except ImportError:
            raise ImportError(self.HF_DOWNLOAD_TEXT)
        if os.path.exists(checkpoint_path) and os.path.exists(
            os.path.join(checkpoint_path, 'config.json')
        ):
            logger.info("HF model's config.json already exists")
            return True
        os.makedirs(checkpoint_path, exist_ok=True)
        try:
            is_hf_model_exists = self.validate_hf_model()
            if not is_hf_model_exists:
                logger.error("Model %s not found on Hugging Face" % (self.repo_id))
                return False
            hf_hub_download(
                repo_id=self.repo_id, filename='config.json', local_dir=checkpoint_path
            )
        except Exception as e:
            logger.error(f"Error downloading model's config.json {e}")
            return False
        return True

    def validate_hf_model(
        self,
    ):
        # check if model exists on HF
        try:
            from huggingface_hub import file_exists, repo_exists
        except ImportError:
            raise ImportError(self.HF_DOWNLOAD_TEXT)
        if self.clarifai_model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE:
            return repo_exists(self.repo_id) and file_exists(self.repo_id, 'config.json')
        else:
            return repo_exists(self.repo_id)

    def get_repo_files_list(self, allowed_file_patterns: list, ignore_file_patterns: list):
        # check if model exists on HF
        try:
            from huggingface_hub import list_repo_files
        except ImportError:
            raise ImportError(self.HF_DOWNLOAD_TEXT)
        # Get the list of files on the repo
        repo_files = list_repo_files(self.repo_id, token=self.token)

        # Get the list of files on the repo that are allowed
        if allowed_file_patterns:

            def should_allow(file_path):
                return any(
                    fnmatch.fnmatch(file_path, pattern) for pattern in allowed_file_patterns
                )

            repo_files = [f for f in repo_files if should_allow(f)]

        self.ignore_patterns = self._get_ignore_patterns()
        if ignore_file_patterns:
            if self.ignore_patterns:
                self.ignore_patterns.extend(ignore_file_patterns)
            else:
                self.ignore_patterns = ignore_file_patterns
        # Get the list of files on the repo that are not ignored
        if getattr(self, "ignore_patterns", None):
            patterns = self.ignore_patterns

            def should_ignore(file_path):
                return any(fnmatch.fnmatch(file_path, pattern) for pattern in patterns)

            repo_files = [f for f in repo_files if not should_ignore(f)]
        return repo_files

    def validate_download(
        self, checkpoint_path: str, allowed_file_patterns: list, ignore_file_patterns: list
    ):
        repo_files = self.get_repo_files_list(
            allowed_file_patterns=allowed_file_patterns, ignore_file_patterns=ignore_file_patterns
        )

        # Check if downloaded files match the files we expect (ignoring ignored patterns)
        checkpoint_dir_files = []
        for dp, dn, fn in os.walk(os.path.expanduser(checkpoint_path)):
            checkpoint_dir_files.extend(
                [os.path.relpath(os.path.join(dp, f), checkpoint_path) for f in fn]
            )

        # Validate by comparing file lists
        return (
            len(checkpoint_dir_files) >= len(repo_files)
            and not (len(set(repo_files) - set(checkpoint_dir_files)) > 0)
            and len(repo_files) > 0
        )

    def _get_ignore_patterns(self):
        # check if model exists on HF
        try:
            from huggingface_hub import list_repo_files
        except ImportError:
            raise ImportError(self.HF_DOWNLOAD_TEXT)

        # Get the list of files on the repo that are not ignored
        repo_files = list_repo_files(self.repo_id, token=self.token)
        self.ignore_patterns = None
        if any(f.endswith(".safetensors") for f in repo_files):
            self.ignore_patterns = [
                "**/original/*",
                "original/*",
                "**/*.pth",
                "**/*.bin",
                "*.pth",
                "*.bin",
                "**/.cache/*",
            ]
        return self.ignore_patterns

    @classmethod
    def validate_hf_repo_access(cls, repo_id: str, token: str = None) -> bool:
        # check if model exists on HF
        try:
            from huggingface_hub import auth_check
            from huggingface_hub.utils import GatedRepoError, RepositoryNotFoundError
        except ImportError:
            raise ImportError(cls.HF_DOWNLOAD_TEXT)

        try:
            auth_check(repo_id, token=token)
            logger.info("Hugging Face repo access validated")
            return True
        except GatedRepoError:
            logger.error(
                "Hugging Face repo is gated. Please make sure you have access to the repo."
            )
            return False
        except RepositoryNotFoundError:
            logger.error("Hugging Face repo not found. Please make sure the repo exists.")
            return False

    @staticmethod
    def validate_config(checkpoint_path: str):
        # check if downloaded config.json exists
        return os.path.exists(checkpoint_path) and os.path.exists(
            os.path.join(checkpoint_path, 'config.json')
        )

    @staticmethod
    def validate_concept(checkpoint_path: str):
        # check if downloaded concept exists in hf model
        config_path = os.path.join(checkpoint_path, 'config.json')
        with open(config_path, 'r') as f:
            config = json.load(f)

        labels = config.get('id2label', None)
        if labels:
            return True
        return False

    @staticmethod
    def fetch_labels(checkpoint_path: str):
        # Fetch labels for classification, detection and segmentation models
        config_path = os.path.join(checkpoint_path, 'config.json')
        with open(config_path, 'r') as f:
            config = json.load(f)

        labels = config['id2label']
        return labels

    @staticmethod
    def get_huggingface_checkpoint_total_size(repo_name, checkpoint_files_list=None):
        """
        Fetches the JSON data for a Hugging Face model using the API with `?blobs=true`.
        Calculates the total size from the JSON output.

        Args:
            repo_name (str): The name of the model on Hugging Face Hub. e.g. "casperhansen/llama-3-8b-instruct-awq"
            checkpoint_files_list (list, optional): A list of specific files to include in the size calculation. If None, all files are included.

        Returns:
            int: The total size in bytes.
        """
        try:
            url = f"https://huggingface.co/api/models/{repo_name}?blobs=true"
            response = requests.get(url)
            response.raise_for_status()  # Raise an exception for bad status codes
            json_data = response.json()

            if isinstance(json_data, str):
                data = json.loads(json_data)
            else:
                data = json_data

            total_size = 0
            for file in data['siblings']:
                if checkpoint_files_list and (file['rfilename'] not in checkpoint_files_list):
                    continue
                total_size += file['size']
            return total_size
        except Exception as e:
            logger.error(f"Error fetching checkpoint size from huggingface.co: {e}")
            return 0



import os
import shutil

import click

from clarifai.cli.base import cli
from clarifai.client.app import App
from clarifai.client.user import User
from clarifai.utils.cli import (
    AliasedGroup,
    convert_timestamp_to_string,
    display_co_resources,
    validate_context,
)
from clarifai.utils.logging import logger


@cli.group(
    ['pipeline', 'pl'],
    cls=AliasedGroup,
    context_settings={'max_content_width': shutil.get_terminal_size().columns - 10},
)
def pipeline():
    """Manage pipelines: upload, init, list, etc"""


@pipeline.command()
@click.argument("path", type=click.Path(exists=True), required=False, default=".")
def upload(path):
    """Upload a pipeline with associated pipeline steps to Clarifai.

    PATH: Path to the pipeline configuration file or directory containing config.yaml. If not specified, the current directory is used by default.
    """
    from clarifai.runners.pipelines.pipeline_builder import upload_pipeline

    upload_pipeline(path)


@pipeline.command()
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=False,
    help='Path to the pipeline run config file.',
)
@click.option('--pipeline_id', required=False, help='Pipeline ID to run.')
@click.option('--pipeline_version_id', required=False, help='Pipeline Version ID to run.')
@click.option(
    '--pipeline_version_run_id',
    required=False,
    help='Pipeline Version Run ID. If not provided, a UUID will be generated.',
)
@click.option('--user_id', required=False, help='User ID of the pipeline.')
@click.option('--app_id', required=False, help='App ID that contains the pipeline.')
@click.option('--nodepool_id', required=False, help='Nodepool ID to run the pipeline on.')
@click.option(
    '--compute_cluster_id', required=False, help='Compute Cluster ID to run the pipeline on.'
)
@click.option('--pipeline_url', required=False, help='Pipeline URL to run.')
@click.option(
    '--timeout',
    type=int,
    default=3600,
    help='Maximum time to wait for completion in seconds. Default 3600 (1 hour).',
)
@click.option(
    '--monitor_interval',
    type=int,
    default=10,
    help='Interval between status checks in seconds. Default 10.',
)
@click.option(
    '--log_file',
    type=click.Path(),
    required=False,
    help='Path to file where logs should be written. If not provided, logs are displayed on console.',
)
@click.option(
    '--monitor',
    is_flag=True,
    default=False,
    help='Monitor an existing pipeline run instead of starting a new one. Requires pipeline_version_run_id.',
)
@click.pass_context
def run(
    ctx,
    config,
    pipeline_id,
    pipeline_version_id,
    pipeline_version_run_id,
    user_id,
    app_id,
    nodepool_id,
    compute_cluster_id,
    pipeline_url,
    timeout,
    monitor_interval,
    log_file,
    monitor,
):
    """Run a pipeline and monitor its progress."""
    import json

    from clarifai.client.pipeline import Pipeline
    from clarifai.utils.cli import from_yaml, validate_context

    validate_context(ctx)

    if config:
        config_data = from_yaml(config)
        pipeline_id = config_data.get('pipeline_id', pipeline_id)
        pipeline_version_id = config_data.get('pipeline_version_id', pipeline_version_id)
        pipeline_version_run_id = config_data.get(
            'pipeline_version_run_id', pipeline_version_run_id
        )
        user_id = config_data.get('user_id', user_id)
        app_id = config_data.get('app_id', app_id)
        nodepool_id = config_data.get('nodepool_id', nodepool_id)
        compute_cluster_id = config_data.get('compute_cluster_id', compute_cluster_id)
        pipeline_url = config_data.get('pipeline_url', pipeline_url)
        timeout = config_data.get('timeout', timeout)
        monitor_interval = config_data.get('monitor_interval', monitor_interval)
        log_file = config_data.get('log_file', log_file)
        monitor = config_data.get('monitor', monitor)
    elif ctx.obj.current:
        if not user_id:
            user_id = ctx.obj.current.get('user_id', '')
        if not app_id:
            app_id = ctx.obj.current.get('app_id', '')
        if not pipeline_id:
            pipeline_id = ctx.obj.current.get('pipeline_id', '')
        if not pipeline_version_id:
            pipeline_version_id = ctx.obj.current.get('pipeline_version_id', '')
        if not nodepool_id:
            nodepool_id = ctx.obj.current.get('nodepool_id', '')
        if not compute_cluster_id:
            compute_cluster_id = ctx.obj.current.get('compute_cluster_id', '')

    # compute_cluster_id and nodepool_id are mandatory regardless of whether pipeline_url is provided
    if not compute_cluster_id or not nodepool_id:
        raise ValueError("--compute_cluster_id and --nodepool_id are mandatory parameters.")

    # When monitor flag is used, pipeline_version_run_id is mandatory
    if monitor and not pipeline_version_run_id:
        raise ValueError("--pipeline_version_run_id is required when using --monitor flag.")

    if pipeline_url:
        # When using pipeline_url, other parameters are optional (will be parsed from URL)
        required_params_provided = True
    else:
        # When not using pipeline_url, all individual parameters are required
        required_params_provided = all([pipeline_id, user_id, app_id, pipeline_version_id])

    if not required_params_provided:
        raise ValueError(
            "Either --user_id & --app_id & --pipeline_id & --pipeline_version_id or --pipeline_url must be provided."
        )

    if pipeline_url:
        pipeline = Pipeline(
            url=pipeline_url,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
            pipeline_version_run_id=pipeline_version_run_id,
            nodepool_id=nodepool_id,
            compute_cluster_id=compute_cluster_id,
            log_file=log_file,
        )
    else:
        pipeline = Pipeline(
            pipeline_id=pipeline_id,
            pipeline_version_id=pipeline_version_id,
            pipeline_version_run_id=pipeline_version_run_id,
            user_id=user_id,
            app_id=app_id,
            nodepool_id=nodepool_id,
            compute_cluster_id=compute_cluster_id,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
            log_file=log_file,
        )

    if monitor:
        # Monitor existing pipeline run instead of starting new one
        result = pipeline.monitor_only(timeout=timeout, monitor_interval=monitor_interval)
    else:
        # Start new pipeline run and monitor it
        result = pipeline.run(timeout=timeout, monitor_interval=monitor_interval)
    click.echo(json.dumps(result, indent=2, default=str))


@pipeline.command()
@click.argument(
    "pipeline_path",
    type=click.Path(),
    required=False,
    default=".",
)
def init(pipeline_path):
    """Initialize a new pipeline project structure.

    Creates the following structure in the specified directory:
    ├── config.yaml          # Pipeline configuration
    ├── stepA/               # First pipeline step
    │   ├── config.yaml     # Step A configuration
    │   ├── requirements.txt # Step A dependencies
    │   └── 1/
    │       └── pipeline_step.py  # Step A implementation
    ├── stepB/               # Second pipeline step
    │   ├── config.yaml     # Step B configuration
    │   ├── requirements.txt # Step B dependencies
    │   └── 1/
    │       └── pipeline_step.py  # Step B implementation
    └── README.md           # Documentation

    PIPELINE_PATH: Path where to create the pipeline project structure. If not specified, the current directory is used by default.
    """
    from clarifai.cli.templates.pipeline_templates import (
        get_pipeline_config_template,
        get_pipeline_step_config_template,
        get_pipeline_step_requirements_template,
        get_pipeline_step_template,
        get_readme_template,
    )

    # Resolve the absolute path
    pipeline_path = os.path.abspath(pipeline_path)

    # Create the pipeline directory if it doesn't exist
    os.makedirs(pipeline_path, exist_ok=True)

    # Prompt for user inputs
    click.echo("Welcome to Clarifai Pipeline Initialization!")
    click.echo("Please provide the following information:")

    user_id = click.prompt("User ID", type=str)
    app_id = click.prompt("App ID", type=str)
    pipeline_id = click.prompt("Pipeline ID", default="hello-world-pipeline", type=str)
    num_steps = click.prompt("Number of pipeline steps", default=2, type=int)

    # Get step names
    step_names = []
    default_names = ["stepA", "stepB", "stepC", "stepD", "stepE", "stepF"]

    for i in range(num_steps):
        default_name = default_names[i] if i < len(default_names) else f"step{i + 1}"
        step_name = click.prompt(f"Name for step {i + 1}", default=default_name, type=str)
        step_names.append(step_name)

    click.echo(f"\nCreating pipeline '{pipeline_id}' with steps: {', '.join(step_names)}")

    # Create pipeline config.yaml
    config_path = os.path.join(pipeline_path, "config.yaml")
    if os.path.exists(config_path):
        logger.warning(f"File {config_path} already exists, skipping...")
    else:
        config_template = get_pipeline_config_template(
            pipeline_id=pipeline_id, user_id=user_id, app_id=app_id, step_names=step_names
        )
        with open(config_path, 'w', encoding='utf-8') as f:
            f.write(config_template)
        logger.info(f"Created {config_path}")

    # Create README.md
    readme_path = os.path.join(pipeline_path, "README.md")
    if os.path.exists(readme_path):
        logger.warning(f"File {readme_path} already exists, skipping...")
    else:
        readme_template = get_readme_template()
        with open(readme_path, 'w', encoding='utf-8') as f:
            f.write(readme_template)
        logger.info(f"Created {readme_path}")

    # Create pipeline steps
    for step_id in step_names:
        step_dir = os.path.join(pipeline_path, step_id)
        os.makedirs(step_dir, exist_ok=True)

        # Create the 1/ subdirectory for the step version
        step_version_dir = os.path.join(step_dir, "1")
        os.makedirs(step_version_dir, exist_ok=True)

        # Create step config.yaml
        step_config_path = os.path.join(step_dir, "config.yaml")
        if os.path.exists(step_config_path):
            logger.warning(f"File {step_config_path} already exists, skipping...")
        else:
            step_config_template = get_pipeline_step_config_template(
                step_id=step_id, user_id=user_id, app_id=app_id
            )
            with open(step_config_path, 'w', encoding='utf-8') as f:
                f.write(step_config_template)
            logger.info(f"Created {step_config_path}")

        # Create step requirements.txt
        step_requirements_path = os.path.join(step_dir, "requirements.txt")
        if os.path.exists(step_requirements_path):
            logger.warning(f"File {step_requirements_path} already exists, skipping...")
        else:
            step_requirements_template = get_pipeline_step_requirements_template()
            with open(step_requirements_path, 'w', encoding='utf-8') as f:
                f.write(step_requirements_template)
            logger.info(f"Created {step_requirements_path}")

        # Create step pipeline_step.py
        step_py_path = os.path.join(step_version_dir, "pipeline_step.py")
        if os.path.exists(step_py_path):
            logger.warning(f"File {step_py_path} already exists, skipping...")
        else:
            step_py_template = get_pipeline_step_template(step_id)
            with open(step_py_path, 'w', encoding='utf-8') as f:
                f.write(step_py_template)
            logger.info(f"Created {step_py_path}")

    logger.info(f"Pipeline initialization complete in {pipeline_path}")
    logger.info("Next steps:")
    logger.info("1. Implement your pipeline step logic in the generated pipeline_step.py files")
    logger.info("2. Add dependencies to requirements.txt files as needed")
    logger.info("3. Run 'clarifai pipeline upload config.yaml' to upload your pipeline")


@pipeline.command(['ls'])
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.option(
    '--app_id',
    required=False,
    help='App ID to list pipelines from. If not provided, lists across all apps.',
)
@click.pass_context
def list(ctx, page_no, per_page, app_id):
    """List all pipelines for the user."""
    validate_context(ctx)

    if app_id:
        app = App(
            app_id=app_id,
            user_id=ctx.obj.current.user_id,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
        )
        response = app.list_pipelines(page_no=page_no, per_page=per_page)
    else:
        user = User(
            user_id=ctx.obj.current.user_id,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
        )
        response = user.list_pipelines(page_no=page_no, per_page=per_page)

    display_co_resources(
        response,
        custom_columns={
            'ID': lambda p: getattr(p, 'pipeline_id', ''),
            'USER_ID': lambda p: getattr(p, 'user_id', ''),
            'APP_ID': lambda p: getattr(p, 'app_id', ''),
            'VERSION_ID': lambda p: getattr(p, 'pipeline_version_id', ''),
            'DESCRIPTION': lambda p: getattr(p, 'description', ''),
            'CREATED_AT': lambda ps: convert_timestamp_to_string(getattr(ps, 'created_at', '')),
            'MODIFIED_AT': lambda ps: convert_timestamp_to_string(getattr(ps, 'modified_at', '')),
        },
        sort_by_columns=[
            ('CREATED_AT', 'desc'),
            ('ID', 'asc'),
        ],
    )



import os

from clarifai import __version__

CLIENT_VERSION = __version__
OS_VER = os.sys.platform
PYTHON_VERSION = '.'.join(
    map(str, [os.sys.version_info.major, os.sys.version_info.minor, os.sys.version_info.micro])
)



from .data_types import JSON  # noqa
from .data_types import Audio  # noqa
from .data_types import Concept  # noqa
from .data_types import Frame  # noqa
from .data_types import Image  # noqa
from .data_types import MessageData  # noqa
from .data_types import NamedFields  # noqa
from .data_types import NamedFieldsMeta  # noqa
from .data_types import Region  # noqa
from .data_types import Text  # noqa
from .data_types import Video  # noqa
from .data_types import cast  # noqa
